import random
try:
    from .dicer import Dicer
except ImportError:
    from dicergirl.utils.dicer import Dicer

class Docimasy:
    """
        检定结果解析类
    """
    judge_dict = {
        "critical success": 4,
        "very hard success": 3,
        "hard success": 2,
        "success": 1,
        "fail": 0,
        "fatal fail": -1,
    }
    def __init__(self, detail, judge: str | int=None):
        self.detail = detail

        if isinstance(judge, int):
            self.judge = judge

        if judge:
            self.judge = self.judge_dict[judge]
        else:
            self.judge = None

    def set_judge(self, judge):
        if isinstance(judge, int):
            self.judge = judge

        self.judge = self.judge_dict[judge]

    def __bool__(self):
        if self.judge > 0:
            return True
        else:
            return False

    def __add__(self, toadd):
        if isinstance(toadd, str):
            toadd = toadd.lstrip("\n")
            if toadd:
                self.detail += "\n" + toadd
        elif isinstance(toadd, int):
            self.judge += toadd
        elif isinstance(toadd, Docimasy):
            self.detail += "\n" + toadd.detail
        else:
            raise NotImplementedError
        return self

    def __int__(self):
        return self.judge

    def __repr__(self):
        return self.detail

    def __str__(self):
        return self.detail

def expr(dice: Dicer, anum) -> Docimasy:
    """ COC 技能检定结果 """
    result = dice.roll().calc()
    docimasy = Docimasy(f"掷骰: {dice.db}")
    docimasy += dice.description()

    if anum:
        if result == 100:
            docimasy += "大失败！"
            docimasy.set_judge("fatal fail")
        elif anum < 50 and result > 95:
            docimasy += f"{result}>95 大失败！"
            docimasy.set_judge("fatal fail")
        elif result == 1:
            docimasy += "大成功！"
            docimasy.set_judge("critical success")
        elif result <= anum // 5:
            docimasy += f"检定值: {anum} {result}≤{anum//5}"
            docimasy += "检定结果: 极难成功."
            docimasy.set_judge("very hard success")
        elif result <= anum // 2:
            docimasy += f"检定值: {anum} {result}≤{anum//2}"
            docimasy += "检定结果: 困难成功."
            docimasy.set_judge("hard success")
        elif result <= anum:
            docimasy += f"检定值: {anum} {result}≤{anum}"
            docimasy += "检定结果: 成功."
            docimasy.set_judge("success")
        else:
            docimasy += f"检定值: {anum} {result}>{anum}"
            docimasy += "检定结果: 失败."
            docimasy.set_judge("fail")
    return docimasy

def scp_doc(result, difficulty, encourage=None, agent=None, great=False) -> Docimasy:
    """ SCP 技能检定结果 """
    if not agent:
        agent = "该特工"

    r = Docimasy(f"事件难度: {difficulty}")

    if difficulty > 25 and result <= difficulty:
        r += f"检定数据: {result}"
        r += f"检定结果: 致命失败."
        r += f"检定结论: {agent} 在试图挑战数学、挑战科学、挑战真理, 尝试达成一个不可能事件, {agent} 毫无疑问获得了 致命失败."
        r.set_judge("fatal fail")
        return r

    if encourage:
        r += f"肾上腺素: {encourage}"
        r += f"检定数据: {result}+{encourage}"
        result += encourage
    else:
        r += f"检定数据: {result}"

    if great:
        r += "检定结果: 关键成功."
        if result <= difficulty:
            r += "检定结论: 有时候, 一次普通的成功或许会大幅度的牵扯到整个未来, 但这并不是一件太过于值得高兴的事情, 因为它不见得是一个好的开始."
        else:
            r += "检定结论: 被 Administrator 所眷顾的人, 毫无疑问这是一次完美的成功, 但是你或许会面对更加绝望的未来."
        r.set_judge("critical success")
    elif result >= (difficulty*2):
        r += "检定结果: 关键成功."
        r += "检定结论: 绝境之中的人常常能够爆发出无尽的潜力, 疯狂是人类最强大的武器, 用疯狂去嗤笑命运吧."
        r.set_judge("critical success")
    elif result > difficulty:
        r += "检定结果: 成功."
        r += "检定结论: 命运常常给予人们无声的嗤笑, 一次成功当然是好事, 但也要警惕这是否是步入深渊的开始."
        r.set_judge("success")
    elif result < (difficulty/2):
        r += "检定结果: 致命失败."
        r += "检定结论: 努力或许的确有用处, 但是努力只是提高运气的一种手段. 在低劣的运气面前, 任何努力都是没有用的."
        r.set_judge("fatal fail")
    elif result < difficulty:
        r += "检定结果: 失败."
        r += "检定结论: 人类从来都生活在饱含恐惧与绝望的危险之中, 失败是一件稀松平常的事情, 小心, 错误的决定或许会让你步入深渊."
        r.set_judge("fail")
    else:
        result = random.randint(0, 1)
        if result:
            r += "检定结果: 成功."
            r += "检定结论: 小心, 你的成功完全是一次偶然, 不要试图去将这样的偶然当做希望."
            r.set_judge("success")
        else:
            r += "检定结果: 失败."
            r += "检定结论: 成功与失败宛如山顶与深渊, 无论是哪一种都是可能的, 相反, 在这个世界落入深渊是一件更加合理的事情."
            r.set_judge("fail")
    return r

def dnd_doc(result, dc, adventurer=None):
    """ DND 技能检定结果 """
    if not adventurer:
        adventurer = "该冒险者"
    r = f"事件难度: {dc}\n"
    r += f"检定数据: {result}\n"
    if result >= 20:
        r += "检定结果: 大成功.\n"
        r += "检定结论: 被命运眷顾的幸运者, 这毫无疑问是一次完美的成功."
    elif result > dc:
        r += "检定结果: 成功.\n"
        r += "检定结论: 前进吧, 冒险者, 异世的诗篇还在等着你."
    elif result <= dc / 2:
        r += "检定结果: 大失败.\n"
        r += "检定结论: 冒险不是自寻死路, 有时候, 放弃也是一个好的选择."
    elif result < dc:
        r += "检定结果: 失败.\n"
        r += "检定结论: 成功与失败总是相辅相成, 不要让一次失败打倒你."
    else:
        result = random.randint(0, 1)
        if result:
            r += "检定结果: 成功.\n"
            r += "检定结论: 小心, 你的成功完全是一次偶然, 不要认为这样的偶然稀松平常, 冷静与冲动并存, 才是一个合格的冒险者."
        else:
            r += "检定结果: 失败.\n"
            r += "检定结论: 成功与失败由于一体两面, 无论是哪一种都是可能的, 但是你不必气馁, 失败与成功都是冒险的一部分."
    return r

if __name__ == "__main__":
    print(scp_doc(10, 20, agent="Oracle"))


from pathlib import Path
from datetime import datetime
from nonebot.plugin import PluginMetadata
from typing import Dict
from .utils.settings import set_package, get_package
from .utils.multilogging import multilogger

import nonebot

__plugin_meta__ = PluginMetadata(
    name="欧若可骰娘",
    description="完善的可拓展跑团机器人, 支持 COC/DND/SCP 等跑团模式.",
    usage="安装即可使用.",
    type="application",
    homepage="https://gitee.com/unvisitor/dicer",
    supported_adapters={"~onebot.v11"},
)
__author__ = "苏向夜 <fu050409@163.com>"

logger = multilogger(name="Dicer Girl", payload="Nonebot2")
try:
    driver = nonebot.get_driver()
    set_package("nonebot2")
except ValueError:
    set_package("qqguild")

from .utils.utils import version as __version__

import logging
import sys
import platform
import psutil
import html

DEBUG = False
current_dir = Path(__file__).resolve().parent
package = get_package()

if package == "nonebot2":
    from .utils.messages import help_message, version
    from .utils.utils import (
        init, on_startswith,
        get_group_id, get_mentions,
        is_super_user, add_super_user, rm_super_user, su_uuid,
        format_msg, format_str,
        get_mode, set_mode,
        get_loggers, loggers, add_logger, remove_logger, log_dir,
        get_status, boton, botoff,
        rolekp, roleob
        )
    from .utils.plugins import modes
    from .utils.parser import CommandParser, Commands, Only, Optional, Required
    from .utils.handlers import show_handler, set_handler, del_handler, roll, shoot
    from .utils.chat import chat

    from nonebot.matcher import Matcher
    from nonebot.plugin import on
    from nonebot.adapters import Bot as Bot
    from nonebot.adapters.onebot.v11 import Bot as V11Bot
    from nonebot.internal.matcher.matcher import Matcher

    if driver._adapters.get("OneBot V12"):
        from nonebot.adapters.onebot.v12 import MessageEvent, GroupMessageEvent, Event, MessageSegment
    else:
        from nonebot.adapters.onebot.v11 import MessageEvent, GroupMessageEvent, Event, MessageSegment

    # 指令装饰器实例化
    testcommand = on_startswith(".test", priority=2, block=True)
    debugcommand = on_startswith(".debug", priority=2, block=True)
    superusercommand = on_startswith((".su", ".sudo"), priority=2, block=True)
    botcommand = on_startswith(".bot", priority=1, block=True)
    logcommand = on_startswith(".log", priority=1, block=True)
    loghandlercommand = on_startswith("", priority=1, block=False)
    selflogcommand = on("message_sent", priority=1, block=False)
    showcommand = on_startswith((".show", ".display"), priority=2, block=True)
    setcommand = on_startswith((".set", ".st"), priority=2, block=True)
    helpcommand = on_startswith((".help", ".h"), priority=2, block=True)
    modecommand = on_startswith((".mode", ".m"), priority=2, block=True)
    shootcommand = on_startswith((".sht", ".shoot"), priority=2, block=True)
    attackcommand = on_startswith((".at", ".attack"), priority=2, block=True)
    damcommand = on_startswith((".dam", ".damage"), priority=2, block=True)
    encommand = on_startswith((".en", ".encourage"), priority=2, block=True)
    racommand = on_startswith(".ra", priority=2, block=True)
    rhcommand = on_startswith(".rh", priority=2, block=True)
    rhacommand = on_startswith(".rha", priority=1, block=True)
    rollcommand = on_startswith((".r", ".roll"), priority=3, block=True)
    delcommand = on_startswith((".del", ".delete"), priority=2, block=True)
    rolekpcommand = on_startswith(".kp", priority=2, block=True)
    roleobcommand = on_startswith(".ob", priority=2, block=True)
    chatcommand = on_startswith(".chat", priority=2, block=True)
    versioncommand = on_startswith((".version", ".v"), priority=2, block=True)

    # 定时任务
    scheduler = nonebot.require("nonebot_plugin_apscheduler").scheduler

    @driver.on_startup
    async def _() -> None:
        """ `Nonebot2`核心加载完成后的初始化方法 """
        global DEBUG
        logger.info("欧若可骰娘初始化中...")
        if DEBUG:
            logging.getLogger().setLevel(logging.DEBUG)
            logger.remove()
            logger.add(
                sys.stdout,
                level = "DEBUG"
            )
            logger.info("DEBUG 模式已启动.")
        init()
        logger.success("欧若可骰娘初始化完毕.")

    @testcommand.handle()
    async def testhandler(matcher: Matcher, event: MessageEvent):
        """ 测试指令 """
        args = format_msg(event.get_message(), begin=".test")
        cp = CommandParser(
            Commands([
                Only("all"),
                Only("split"),
                Only("markdown")
            ]),
            args=args,
            auto=True
        )

        if not is_super_user(event):
            await matcher.send("[Oracle] 权限不足, 拒绝执行测试指令.")
            return

        reply = ""
        if cp.nothing or cp.results["all"]:
            import json
            reply += f"收到消息: {event.get_message()}\n"
            reply += f"消息来源: {event.group_id} - {event.get_user_id()}\n"
            reply += f"消息拆析: {args}\n"
            reply += f"指令拆析: {cp.results}\n"
            reply += f"消息原始内容: {event.get_plaintext()}\n"
            reply += f"消息json数据: {event.json()}\n"
            reply += f"消息发送者json信息: {json.loads(event.json())['sender']}\n"
            reply += f"发送者昵称: {json.loads(event.json())['sender']['nickname']}"
            await matcher.send(reply)
            return

        logger.debug("收到消息:" + str(event.get_message()))

        if not msg:
            msg = "[]"

        if msg[-1] == "markdown":
            mp = ""
            await matcher.send(group_id=event.group_id, message=mp)
            return

        await matcher.send(None)

    @debugcommand.handle()
    async def debughandler(matcher: Matcher, event: MessageEvent):
        """ 漏洞检测指令 """
        global DEBUG
        args = format_msg(event.get_message(), begin=".debug")
        if not is_super_user(event):
            await matcher.send("[Oracle] 权限不足, 拒绝执行指令.")
            return

        if args:
            logger.debug(args)
            if args[0] == "off":
                DEBUG = False
                logging.getLogger().setLevel(logging.INFO)
                logger.remove()
                logger.add(
                    sys.stdout,
                    level = "INFO"
                )
                logger.info("[cocdicer] 输出等级设置为 INFO.")
                await matcher.send("[Oracle] DEBUG 模式已关闭.")
                return
        else:
            DEBUG = True
            logging.getLogger().setLevel(logging.DEBUG)
            logger.remove()
            logger.add(
                sys.stdout,
                level = "INFO"
            )
            logger.info("[cocdicer] 输出等级设置为 DEBUG.")
            await matcher.send("[Oracle] DEBUG 模式已启动.")
            return

        if args[0] == "on":
            DEBUG = True
            logging.getLogger().setLevel(logging.DEBUG)
            logger.remove()
            logger.add(
                sys.stdout,
                level = "INFO"
            )
            logger.info("[cocdicer] 输出等级设置为 DEBUG.")
            await matcher.send("[Oracle] DEBUG 模式已启动.")
        else:
            await matcher.send("[Oracle] 错误, 我无法解析你的指令.")

    @superusercommand.handle()
    async def superuser_handler(matcher: Matcher, event: MessageEvent):
        """ 超级用户管理指令 """
        args = format_str(event.get_message(), begin=(".su", ".sudo"))
        arg = list(filter(None, args.split(" ")))

        if len(arg) >= 1:
            if arg[0].lower() == "exit":
                if not rm_super_user(event):
                    await matcher.send("[Oracle] 你还不是超级管理员, 无法撤销超级管理员身份.")
                    return
                await matcher.send("[Oracle] 你已撤销超级管理员身份.")
                return

        if is_super_user(event):
            await matcher.send("[Oracle] 你已经是超级管理员.")
            return

        if not args:
            logger.critical(f"超级令牌: {su_uuid}")
            await matcher.send("[Oracle] 启动超级管理员鉴权, 鉴权令牌已在控制终端展示.")
        else:
            if not args == su_uuid:
                await matcher.send("[Oracle] 鉴权失败!")
            else:
                add_super_user(event)
                await matcher.send("[Oracle] 你取得了管理员权限.")

    @botcommand.handle()
    async def bothandler(bot: V11Bot, matcher: Matcher, event: GroupMessageEvent):
        """ 机器人管理指令 """
        args = format_msg(event.get_message(), begin=".bot")
        commands = CommandParser(
            Commands([
                Only(("version", "v", "bot")),
                Only(("exit", "bye", "leave")),
                Only(("on", "run", "start")),
                Only(("off", "down", "shutdown")),
                Only(("status"))
            ]),
            args=args,
            auto=True
        )

        if commands.nothing or commands.results["version"]:
            return await versionhandler(matcher=matcher)
        else:
            commands = commands.results

        if commands["exit"]:
            if not is_super_user(event):
                await matcher.send("[Oracle] 你没有管理员权限, 请先执行`.su`开启权限鉴定.")
                return

            logger.info(f"欧若可退出群聊: {event.group_id}")
            await matcher.send("[Oracle] 欧若可离开群聊.")
            await bot.set_group_leave(group_id=event.group_id)
            return

        if commands["on"]:
            boton(event)
            await matcher.send("[Oracle] 欧若可已开放指令限制.")
            return

        if commands["off"]:
            botoff(event)
            await matcher.send("[Oracle] 欧若可已开启指令限制.")
            return

        if commands["status"]:
            try:
                system = platform.freedesktop_os_release()["PRETTY_NAME"]
            except:
                system = platform.platform()

            memi = psutil.Process().memory_info()
            rss = memi.rss / 1024 / 1024
            total = psutil.virtual_memory().total / 1024 / 1024

            reply = f"欧若可骰娘 {version}, {'正常运行' if get_status(event) else '指令限制'}\n"
            reply += f"操作系统: {system}\n"
            reply += f"CPU 核心: {psutil.cpu_count()} 核心\n"
            reply += f"Python 版本: {platform.python_version()}\n"
            reply += "系统内存占用: %.2fMB/%.2fMB\n" % (rss, total)
            reply += f"漏洞检测模式: {'on' if DEBUG else 'off'}"
            await matcher.send(reply)
            return

        await matcher.send("[Oracle] 未知的指令, 使用`.help bot`获得机器人管理指令使用帮助.")

    @logcommand.handle()
    async def loghandler(bot: V11Bot, matcher: Matcher, event: Event):
        """ 日志管理指令 """
        args = format_msg(event.get_message(), begin=".log")
        commands = CommandParser(
            Commands([
                Only("show"),
                Only(("add", "new")),
                Optional("name", str),
                Optional("stop", int),
                Optional("start", int),
                Optional(("remove", "rm"), int),
                Optional(("download", "load"), int)
                ]),
            args=args,
            auto=True
            ).results

        if commands["show"]:
            gl = get_loggers(event)
            if len(gl) == 0:
                await matcher.send("暂无存储的日志.")
                return

            if not get_group_id(event) in loggers.keys():
                running = []
            else:
                running = loggers[get_group_id(event)].keys()

            reply = "该群聊所有日志:\n"
            for l in gl:
                index = gl.index(l)
                reply += f"序列 {index} : {l} : {'记录中' if index in running else '已关闭'}\n"
            reply.strip("\n")

            await matcher.send(reply)
            return

        if commands["download"] or commands["download"] == 0:
            gl = get_loggers(event)
            if commands["download"] > len(gl)-1:
                await matcher.send(f"目标日志序列 {commands['download']} 不存在.")
                return

            path = Path(gl[commands["download"]])
            await bot.call_api(
                "upload_group_file",
                **{
                    "group_id": get_group_id(event),
                    "file": str(path),
                    "name": path.name
                }
            )
            return

        if commands["add"]:
            if commands["name"]:
                logname = str(log_dir/(commands["name"]+".trpg.log"))
            else:
                logname = str(log_dir/(datetime.now().strftime("%Y-%m-%d-%H-%M-%S")+".trpg.log"))

            new_logger = multilogger(name="DG Msg Logger", payload="TRPG")
            new_logger.remove()
            new_logger.add(logname, encoding='utf-8', format="{message}", level="INFO")
            if not get_group_id(event) in loggers.keys():
                loggers[get_group_id(event)] = {}

            index = len(get_loggers(event))
            loggers[get_group_id(event)][index] = [new_logger, logname]

            if not add_logger(event, logname):
                raise IOError("无法新增日志.")

            await matcher.send(f"[Oracle] 新增日志序列: {index}\n日志文件: {logname}")
            return

        if commands["stop"] or commands["stop"] == 0:
            if not get_group_id(event) in loggers.keys():
                await matcher.send("该群组无正在运行中的日志.")
                loggers[get_group_id(event)] = {}
                return

            if not commands["stop"] in loggers[get_group_id(event)]:
                await matcher.send("目标日志不存在.")
                return

            loggers[get_group_id(event)][commands["stop"]][0].remove()
            loggers[get_group_id(event)].pop(commands["stop"])
            await matcher.send(f"日志序列 {commands['stop']} 已停止记录.")
            return

        if commands["start"] or commands["start"] == 0:
            gl = get_loggers(event)
            if commands["start"] > len(gl)-1:
                await matcher.send(f"目标日志序列 {commands['start']} 不存在.")
                return

            logname = gl[commands["start"]]
            new_logger = multilogger(name="DG Msg Logger", payload="TRPG")
            new_logger.remove()
            new_logger.add(logname, encoding='utf-8', format="{message}", level="INFO")
            if not get_group_id(event) in loggers.keys():
                loggers[get_group_id(event)] = {}

            loggers[get_group_id(event)][commands["start"]] = [new_logger, logname]
            await matcher.send(f"[Oracle] 日志序列 {commands['start']} 重新启动.")
            return

        if commands["remove"] or commands["remove"] == 0:
            gl = get_loggers(event)
            if not gl:
                await matcher.send("该群组从未设置过日志.")
                loggers[get_group_id(event)] = {}
                return

            if commands["remove"] > len(gl)-1:
                await matcher.send(f"目标日志序列 {commands['remove']} 不存在.")
                return

            index = len(gl)
            if get_group_id(event) in loggers.keys():
                if commands["remove"] in loggers[get_group_id(event)].keys():
                    await matcher.send(f"目标日志序列 {commands['remove']} 正在运行, 开始中止...")
                    loggers[get_group_id(event)][commands["remove"]][0].remove()
                    loggers[get_group_id(event)].pop(commands["remove"])

            Path(gl[commands["remove"]]).unlink()
            remove_logger(event, commands["remove"])
            await matcher.send(f"日志序列 {commands['remove']} 已删除.")
            return

        await matcher.send("骰娘日志管理系统, 使用`.help log`指令详细信息.")

    def trpg_log(event):
        """ 外置的日志记录方法 """
        import json
        if not get_group_id(event) in loggers.keys():
            return
        for log in loggers[get_group_id(event)].keys():
            if isinstance(event, GroupMessageEvent):
                raw_json = json.loads(event.json())
                if raw_json['sender']['card']:
                    if raw_json['sender']['card'].lower() == "ob":
                        role_or_name = f"[旁观者 - {raw_json['sender']['nickname']}]"
                    elif raw_json['sender']['card'].lower() == "kp":
                        role_or_name = f"[主持人 - {raw_json['sender']['nickname']}]"
                    else:
                        role_or_name = f"[{raw_json['sender']['card']}]"
                elif raw_json['sender']['nickname']:
                    role_or_name = f"[未知访客 - {raw_json['sender']['nickname']}]"
                else:
                    role_or_name = f"[未知访客 - {str(event.get_user_id())}]"

                message = role_or_name + ": " + html.unescape(str(event.get_message()))
            elif isinstance(event, Event):
                message = "[欧若可]: " + html.unescape(event.message)

            loggers[get_group_id(event)][log][0].info(message)

    @selflogcommand.handle()
    @loghandlercommand.handle()
    def loggerhandler(event: Event):
        """ 消息记录日志指令 """
        trpg_log(event)

    @showcommand.handle()
    async def showhandler(matcher: Matcher, event: GroupMessageEvent, args: list=None):
        """ 角色卡展示指令 """
        if not get_status(event) and not event.to_me:
            return

        if not isinstance(args, list):
            args = format_msg(event.get_message(), begin=(".show", ".display"))
        else:
            args = args

        at = get_mentions(event)

        mode = get_mode(event)
        if mode in modes:
            try:
                sh = show_handler(event, args, at, mode=mode)
            except Exception as error:
                logger.exception(error)
                sh = [f"[Oracle] 错误: 执行指令失败, 疑似该模式不存在该指令."]
        else:
            await matcher.send("未知的跑团模式.")
            return True

        for msg in sh:
            await matcher.send(str(msg))
        return

    @setcommand.handle()
    async def sethandler(matcher: Matcher, event: GroupMessageEvent):
        """ 角色卡设置指令 """
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=(".set", ".st"))
        at = get_mentions(event)
        commands = CommandParser(
            Commands([
                Only("show"),
                Only("del"),
                Only("clear")
            ]),
            args,
            auto=True
        ).results

        if at and not is_super_user(event):
            await matcher.send("[Oracle] 权限不足, 拒绝执行指令.")
            return

        if commands["show"]:
            args.remove("show")
            return await showhandler(matcher, event, args=args)

        if commands["del"]:
            args.remove("del")
            return await delhandler(matcher, event, args=args)

        try:
            sh = set_handler(event, args, at, mode=mode)
        except Exception as error:
            logger.exception(error)
            sh = [f"[Oracle] 错误: 执行指令失败, 疑似模式 {mode} 不存在该指令."]

        await matcher.send(sh)
        return

    @helpcommand.handle()
    async def helphandler(matcher: Matcher, event: MessageEvent):
        """ 帮助指令 """
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=(".help", ".h"))
        if args:
            arg = args[0]
        else:
            arg = ""
        print(arg)
        await matcher.send(help_message(arg))

    @modecommand.handle()
    async def modehandler(matcher: Matcher, event: MessageEvent):
        """ 跑团模式切换指令 """
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=(".mode", ".m"))
        if args:
            if args[0].lower() in modes:
                set_mode(event, args[0].lower())
                await matcher.send(f"[Oracle] 已切换到 {args[0].upper()} 跑团模式.")
                return True
            else:
                await matcher.send("[Oracle] 未知的跑团模式, 忽略指令.")
                return True
        else:
            reply = "当前已正确安装的跑团插件:\n"
            for plugin in modes.keys():
                reply += f"{plugin.upper()} 模式: {plugin}.\n"

            reply += f"[Oracle] 当前的跑团模式为 {get_mode(event).upper()}."
            await matcher.send(reply)

    @shootcommand.handle()
    async def shoothandler(matcher: Matcher, event: GroupMessageEvent):
        """ 射击检定指令 """
        if not get_status(event) and not event.to_me:
            return

        await matcher.send(shoot())

    @attackcommand.handle()
    async def attackhandler(matcher: Matcher, event: GroupMessageEvent):
        """ 伤害检定指令 """
        if not get_status(event) and not event.to_me:
            return

        args = format_str(event.get_message(), begin=(".at", ".attack"))
        mode = get_mode(event)
        if mode in modes:
            if not hasattr(modes[mode], "__commands__"):
                await matcher.send(f"[Oracle] 跑团模式 {mode.upper()} 未设置标准指令.")
                return

            if not "at" in modes[mode].__commands__.keys():
                await matcher.send(f"[Oracle] 跑团模式 {mode.upper()} 不支持伤害检定指令.")
                return

            handler = modes[mode].__commands__["at"]
            await matcher.send(handler(event, args))
        else:
            await matcher.send("[Oracle] 未知的跑团模式.")

    @damcommand.handle()
    async def damhandler(matcher: Matcher, event: GroupMessageEvent):
        """ 承伤检定指令 """
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=(".dam", ".damage"))
        mode = get_mode(event)
        if mode in modes:
            if not hasattr(modes[mode], "__commands__"):
                await matcher.send(f"[Oracle] 跑团模式 {mode.upper()} 未设置标准指令.")
                return

            if not "at" in modes[mode].__commands__.keys():
                await matcher.send(f"[Oracle] 跑团模式 {mode.upper()} 不支持承伤检定指令.")
                return

            handler = modes[mode].__commands__["dam"]
            await matcher.send(handler(event, args))
        else:
            await matcher.send("[Oracle] 未知的跑团模式.")

    @encommand.handle()
    async def enhandler(matcher: Matcher, event: GroupMessageEvent):
        """ 属性或技能激励指令 """
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=".en")
        mode = get_mode(event)
        if mode in modes:
            if not hasattr(modes[mode], "__commands__"):
                await matcher.send(f"[Oracle] 跑团模式 {mode.upper()} 未设置标准指令.")
                return

            if not "at" in modes[mode].__commands__.keys():
                await matcher.send(f"[Oracle] 跑团模式 {mode.upper()} 不支持激励指令.")
                return

            handler = modes[mode].__commands__["en"]
            await matcher.send(handler(event, args))
        else:
            await matcher.send("[Oracle] 未知的跑团模式.")

    @racommand.handle()
    async def rahandler(matcher: Matcher, event: GroupMessageEvent):
        """ 属性或技能检定指令 """
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=".ra")
        mode = get_mode(event)
        if mode in modes:
            if not hasattr(modes[mode], "__commands__"):
                await matcher.send(f"[Oracle] 跑团模式 {mode.upper()} 未设置标准指令.")
                return

            if not "at" in modes[mode].__commands__.keys():
                await matcher.send(f"[Oracle] 跑团模式 {mode.upper()} 不支持激励指令.")
                return

            handler = modes[mode].__commands__["ra"]
            replies = handler(event, args)
            if isinstance(replies, list):
                for reply in replies:
                    await matcher.send(reply)
                return

            await matcher.send(replies)
        else:
            await matcher.send("[Oracle] 未知的跑团模式.")

    @rhcommand.handle()
    async def rhhandler(bot: Bot, matcher: Matcher, event: GroupMessageEvent):
        """ 暗骰指令 """
        if not get_status(event) and not event.to_me:
            return

        args = format_str(event.get_message(), begin=".rh")
        await matcher.send("[Oracle] 暗骰: 命运的骰子在滚动.")
        await bot.send_private_msg(user_id=event.get_user_id(), message=roll(args))

    @rhacommand.handle()
    async def rhahandler(bot: Bot, matcher: Matcher, event: GroupMessageEvent):
        """ 暗骰技能检定指令 """
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=".rha")
        await matcher.send("[Oracle] 暗骰: 命运的骰子在滚动.")
        await bot.send_private_msg(user_id=event.get_user_id(), message=eval(f"{mode}_(args, event)"))

    @rollcommand.handle()
    async def rollhandler(matcher: Matcher, event: MessageEvent):
        """ 标准掷骰指令 """
        if not get_status(event) and not event.to_me:
            return

        args = format_str(event.get_message(), begin=(".r", ".roll"))
        if not args:
            await matcher.send(roll("1d100"))
            return

        try:
            await matcher.send(roll(args))
        except Exception as error:
            logger.exception(error)
            await matcher.send("[Oracle] 未知错误, 可能是掷骰语法异常.\nBUG提交: https://gitee.com/unvisitor/issues")

    @delcommand.handle()
    async def delhandler(matcher: Matcher, event: GroupMessageEvent, args: list=None):
        """ 角色卡或角色卡技能删除指令 """
        if not get_status(event) and not event.to_me:
            return

        if not isinstance(args, list):
            args = format_msg(event.get_message(), begin=(".del", ".delete"))
        else:
            args = args

        at = get_mentions(event)

        if at and not is_super_user(event):
            await matcher.send("[Oracle] 权限不足, 拒绝执行指令.")
            return

        mode = get_mode(event)
        if mode in modes:
            for msg in del_handler(event, args, at, mode=mode):
                await matcher.send(msg)

    @rolekpcommand.handle()
    async def rolekphandler(bot: V11Bot, matcher: Matcher, event: GroupMessageEvent):
        """ KP 身份组认证 """
        args = format_msg(event.get_message(), begin=(".kp"))
        cp = CommandParser(
            Commands([
                Optional(("time", "schedule"), int),
                Optional(("minute", "min"), int, 0)
            ]),
            args=args,
            auto=True
        ).results

        if cp["time"]:
            @scheduler.scheduled_job("cron", hour=cp["time"], minute=cp["minute"], id="timer")
            async def _():
                await bot.send_group_msg(group_id=event.group_id, message="抵达 KP 设定的跑团启动时间.")

            try:
                scheduler.start()
            except:
                pass
            await matcher.send(f"定时任务: {cp['time']}: {cp['minute'] if len(str(cp['minute'])) > 1 else '0'+str(cp['minute'])}")
            return

        rolekp(event)
        await bot.set_group_card(group_id=event.group_id, user_id=event.get_user_id(), card="KP")
        await matcher.send("[Oracle] 身份组设置为主持人 (KP).")

    @roleobcommand.handle()
    async def roleobcommand(bot: V11Bot, matcher: Matcher, event: GroupMessageEvent):
        """ OB 身份组认证 """
        import json
        roleob(event)

        if json.loads(event.json())['sender']['card'] == "ob":
            await bot.set_group_card(group_id=event.group_id, user_id=event.get_user_id())
            await matcher.send("[Oracle] 取消旁观者 (OB) 身份.")
        else:
            await bot.set_group_card(group_id=event.group_id, user_id=event.get_user_id(), card="ob")
            await matcher.send("[Oracle] 身份组设置为旁观者 (OB).")

    @chatcommand.handle()
    async def chathandler(matcher: Matcher, event: MessageEvent):
        """ chatGPT 对话指令 """
        args = format_str(event.get_message(), begin=".chat")
        if not args:
            await matcher.send("[Oracle] 空消息是不被允许的.")
            return
        await matcher.send(chat(args))

    @versioncommand.handle()
    async def versionhandler(matcher: Matcher):
        """ 骰娘版本及开源声明指令 """
        await matcher.send(f"欧若可骰娘 版本 {version}, 未知访客开发, 以Apache-2.0协议开源.\nCopyright © 2011-2023 Unknown Visitor. Open source as protocol Apache-2.0.")
        return
elif package == "qqguild":
    pass
else:
    logger.critical(f"未知的包模式: {package}!")
    sys.exit()


from typing import Dict, List
try:
    from ..utils.cards import Cards
except ImportError:
    from dicergirl.utils.cards import Cards

dnd_cards = Cards(mode="dnd")
dnd_cache_cards = Cards(mode="dnd")
dnd_attrs_dict: Dict[str, List[str]] = {
    "名字": ["name", "名字", "名称", "姓名"],
    "性别": ["sex", "性别"],
    "年龄": ["age", "年龄"],
    "力量": ["str", "力量", "攻击", "攻击力"],
    "敏捷": ["dex", "敏捷"],
    "体质": ["con", "体质"],
    "智力": ["int", "智力", "灵感"],
    "感知": ["fel", "感知", "感觉", "侦查"],
    "魅力": ["chr", "魅力", "外貌"],
    "生命": ["hp", "生命"]
}



from pathlib import Path
from loguru import logger

import os

def check_subpackages_for_plugin(root_package: Path | str):
    subpackage_names = []
    for root, _, files in os.walk(root_package):
        if "__init__.py" in files:
            package_name = os.path.basename(root)
            subpackage_names.append(package_name)

    plugin_subpackages = []
    for subpackage_name in subpackage_names:
        subpackage_module = subpackage_name #f"{str(root_package).split('.')[-1]}.{subpackage_name}"
        print(subpackage_module)
        try:
            subpackage = __import__(subpackage_module, fromlist=['__type__'])
            if hasattr(subpackage, '__type__'):
                if subpackage.__type__ == "plugin":
                    plugin_subpackages.append(subpackage_name)
        except ImportError as error:
            logger.exception(error)

    return plugin_subpackages

if __name__ == "__main__":
    root_package = Path(__file__).resolve().parent  # Replace with your root package name
    plugin_subpackages = check_subpackages_for_plugin(root_package)
    print("Plugin subpackages:", plugin_subpackages)



from botpy.message import Message
from botpy.logging import get_logger
from pathlib import Path
from utils.settings import set_package, get_package
from multilogging import multilogger
set_package("qqguild")

from coc.investigator import Investigator
from coc.coccards import coc_cards, coc_cache_cards
from coc.cocutils import sc, st, at, coc_dam, coc_en, ra, ti, li, rb, rp

from scp.agent import Agent
from scp.scpcards import scp_cards, scp_cache_cards
from scp.scputils import sra, scp_dam, scp_en, at as sat, deal, begin
from scp.attributes import all_alias_dict

from dnd.adventurer import Adventurer
from dnd.dndcards import dnd_cards, dnd_cache_cards
from dnd.dndutils import dra

from utils.decorators import Commands, translate_punctuation
from utils.messages import help_message, version
from utils.utils import init, is_super_user, add_super_user, rm_super_user, su_uuid, format_msg, format_str, get_handlers, get_config, modes, get_mentions
from utils.handlers import show_handler, set_handler, del_handler, roll
from utils.chat import chat

import botpy
import logging
import sys
import logging
import sys
import random
import re
import asyncio
import nonebot

logger = multilogger(name="Dicer Girl", payload="Nonebot2")
try:
    driver = nonebot.get_driver()
    set_package("nonebot2")
    from nonebot.rule import Rule
    from nonebot.matcher import Matcher
    from nonebot.plugin import on_startswith, on_message
    from nonebot.adapters import Bot as Bot
    from nonebot.adapters.onebot.v11 import Bot as V11Bot
    from nonebot.consts import STARTSWITH_KEY
    if driver._adapters.get("OneBot V12"):
        from nonebot.adapters.onebot.v12 import MessageEvent, GroupMessageEvent
    else:
        from nonebot.adapters.onebot.v11 import MessageEvent, GroupMessageEvent
except ValueError:
    set_package("qqguild")
    class GroupMessageEvent:
        ...
    class Matcher:
        ...
    class V11Bot:
        ...
    class Bot:
        ...

DEBUG = False
current_dir = Path(__file__).resolve().parent
mode = "scp"
package = get_package()
config = get_config()

class StartswithRule:
    __slots__ = ("msg", "ignorecase")

    def __init__(self, msg, ignorecase=False):
        self.msg = msg
        self.ignorecase = ignorecase

    def __repr__(self) -> str:
        return f"Startswith(msg={self.msg}, ignorecase={self.ignorecase})"

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, StartswithRule)
            and frozenset(self.msg) == frozenset(other.msg)
            and self.ignorecase == other.ignorecase
        )

    def __hash__(self) -> int:
        return hash((frozenset(self.msg), self.ignorecase))

    async def __call__(self, event, state) -> bool:
        try:
            text = translate_punctuation(event.get_plaintext())
        except Exception:
            return False
        if match := re.match(
            f"^(?:{'|'.join(re.escape(prefix) for prefix in self.msg)})",
            text,
            re.IGNORECASE if self.ignorecase else 0,
        ):
            state[STARTSWITH_KEY] = match.group()
            return True
        return False

def startswith(msg, ignorecase=True):
    if isinstance(msg, str):
        msg = (msg,)

    return Rule(StartswithRule(msg, ignorecase))

def on_startswith(commands, priority=0, block=True):
    if isinstance(commands, str):
        commands = (commands, )

    if get_package() == "nonebot2":
        return on_message(startswith(commands, True), priority=priority, block=block, _depth=1)
    elif get_package() == "qqguild":
        return Commands(name=commands)

testcommand = on_startswith(".test", priority=2, block=True)
debugcommand = on_startswith(".debug", priority=2, block=True)
superusercommand = on_startswith((".su", ".sudo"), priority=2, block=True)
botcommand = on_startswith(".bot", priority=1, block=True)
coccommand = on_startswith(".coc", priority=1, block=True)
scpcommand = on_startswith(".scp", priority=1, block=True)
dndcommand = on_startswith(".dnd", priority=1, block=True)
showcommand = on_startswith((".show", ".display"), priority=2, block=True)
setcommand = on_startswith((".set", ".st"), priority=2, block=True)
helpcommand = on_startswith((".help", ".h"), priority=2, block=True)
modecommand = on_startswith((".mode", ".m"), priority=2, block=True)
stcommand = on_startswith(".sht", priority=2, block=True)
attackcommand = on_startswith((".at", ".attack"), priority=2, block=True)
damcommand = on_startswith((".dam", ".damage"), priority=2, block=True)
encommand = on_startswith((".en", ".encourage"), priority=2, block=True)
racommand = on_startswith(".ra", priority=2, block=True)
rhcommand = on_startswith(".rh", priority=2, block=True)
rhacommand = on_startswith(".rha", priority=1, block=True)
rcommand = on_startswith((".r", ".roll"), priority=3, block=True)
ticommand = on_startswith(".ti", priority=2, block=True)
licommand = on_startswith(".li", priority=2, block=True)
sccommand = on_startswith(".sc", priority=2, block=True)
delcommand = on_startswith((".del", ".delete"), priority=2, block=True)
chatcommand = on_startswith(".chat", priority=2, block=True)
versioncommand = on_startswith((".version", ".v"), priority=2, block=True)

if get_package() == "nonebot2":
    @driver.on_startup
    async def _():
        global DEBUG
        logger.info("欧若可骰娘初始化中...")
        if DEBUG:
            logging.getLogger().setLevel(logging.DEBUG)
            logger.remove()
            logger.add(
                sys.stdout,
                level = "DEBUG"
            )
            logger.info("DEBUG 模式已启动.")
        init()
        coc_cards.load()
        scp_cards.load()
        logger.success("欧若可骰娘初始化完毕.")

@testcommand.handle()
async def testhandler(matcher: Matcher, event: GroupMessageEvent):
    if not is_super_user(event):
        await matcher.send("[Oracle] 权限不足, 拒绝执行测试指令.")
        return

    logger.info("收到消息:" + str(event.get_message()))
    msg = format_msg(event.get_message(), begin=".test")

    if not msg:
        msg = "[]"

    if msg[-1] == "markdown":
        mp = ""
        await matcher.send(group_id=event.group_id, message=mp)
        return

    await matcher.finish(str(msg))


@debugcommand.handle()
async def debughandler(matcher: Matcher, event: GroupMessageEvent):
    global DEBUG
    args = format_msg(event.get_message(), begin=".debug")
    if not is_super_user(event):
        await matcher.send("[Oracle] 权限不足, 拒绝执行指令.")
        return

    if args:
        logger.debug(args)
        if args[0] == "off":
            DEBUG = False
            logging.getLogger().setLevel(logging.INFO)
            logger.remove()
            logger.add(
                sys.stdout,
                level = "INFO"
            )
            logger.info("[cocdicer] 输出等级设置为 INFO.")
            await matcher.send("[Oracle] DEBUG 模式已关闭.")
            return
    else:
        DEBUG = True
        logging.getLogger().setLevel(logging.DEBUG)
        logger.remove()
        logger.add(
            sys.stdout,
            level = "INFO"
        )
        logger.info("[cocdicer] 输出等级设置为 DEBUG.")
        await matcher.send("[Oracle] DEBUG 模式已启动.")
        return

    if args[0] == "on":
        DEBUG = True
        logging.getLogger().setLevel(logging.DEBUG)
        logger.remove()
        logger.add(
            sys.stdout,
            level = "INFO"
        )
        logger.info("[cocdicer] 输出等级设置为 DEBUG.")
        await matcher.send("[Oracle] DEBUG 模式已启动.")
    else:
        await matcher.send("[Oracle] 错误, 我无法解析你的指令.")

@superusercommand.handle()
async def superuser_handler(matcher: Matcher, event: GroupMessageEvent):
    args = format_str(event.get_message(), begin=(".su", ".sudo"))
    arg = list(filter(None, args.split(" ")))

    if len(arg) >= 1:
        if arg[0].lower() == "exit":
            if not rm_super_user(event):
                await matcher.send("[Oracle] 你还不是超级管理员, 无法撤销超级管理员身份.")
                return
            await matcher.send("[Oracle] 你已撤销超级管理员身份.")
            return

    if is_super_user(event):
        await matcher.send("[Oracle] 你已经是超级管理员.")
        return

    if not args:
        logger.critical(f"超级令牌: {su_uuid}")
        await matcher.send("[Oracle] 启动超级管理员鉴权, 鉴权令牌已在控制终端展示.")
    else:
        if not args == su_uuid:
            await matcher.send("[Oracle] 鉴权失败!")
        else:
            add_super_user(event)
            await matcher.send("[Oracle] 你取得了管理员权限.")

@botcommand.handle()
async def bothandler(bot: V11Bot, matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(event.get_message(), begin=".bot")
    if not is_super_user(event):
        await matcher.send("[Oracle] 你没有管理员权限, 请先执行`.su`开启权限鉴定.")
        return
    if len(args) == 1:
        if args[0] in ["exit", "out", "leave"]:
            print("退出群聊.")
            await matcher.send("[Oracle] 欧若可离开群聊.")
            await bot.set_group_leave(group_id=event.group_id)
        elif args[0] in ["on", "run", "start"]:
            await matcher.send("[Oracle] 我运行在非 systemd 平台上, 我将保持启动.")
        elif args[0] in ["off", "down", "shutdown"]:
            await matcher.send("[Oracle] 我运行在非 systemd 平台上, 我将保持启动.")
        else:
            await matcher.send("[Oracle] 错误的指令.")
    else:
        await matcher.send(help_message("bot"))

@coccommand.handle()
async def cochandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(event.get_message(), begin=".coc")
    if len(args) > 1:
        logger.info("指令错误, 驳回.")
        await matcher.send("[Oracle] 错误: 参数超出预计(1需要 但 %d传入), 指令驳回." % len(args))
        return False

    try:
        if len(args) == 0:
            raise ValueError
        args = int(args[0])
    except ValueError:
        await matcher.send(f'警告: 参数 {args} 不合法, 使用默认值 20 替代.')
        args = 20

    inv = Investigator()
    await matcher.send(inv.age_change(args))

    if 15 <= args and args < 90:
        coc_cache_cards.update(event, inv.__dict__, save=False)
        await matcher.send(str(inv.output()))

@scpcommand.handle()
async def scp_handler(matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(event.get_message(), begin=".scp", zh_en=True)
    at = get_mentions(event)

    if at and not is_super_user(event):
        return "[Oracle] 权限不足, 无法指定玩家修改人物卡."

    if at:
        qid = at[0]
    else:
        qid = ""

    if len(args) != 0:
        if args[0] in ["begin", "start"]:
            for des in begin():
                await matcher.send(des)
                await asyncio.sleep(2)
            return
        elif args[0] in ["reset", "r"]:
            if not is_super_user(event):
                await matcher.send("[Oracle] 权限不足, 拒绝执行人物卡重置指令.")
                return

            agt = Agent().load(scp_cards.get(event, qid=qid))

            if len(args) == 2:
                try:
                    exec(f"agt.reset_{args[1]}()")
                    scp_cards.update(event, agt.__dict__, qid=qid, save=True)
                    await matcher.send(f"[Oracle] 已重置指定人物卡属性: {args[1]}.")
                except:
                    await matcher.send("[Oracle] 指令看起来不存在.")
                finally:
                    return

            agt.reset()
            scp_cards.update(event, agt.__dict__, qid=qid, save=True)
            await matcher.send(f"[Oracle] 人物卡 {agt.name} 属性已重置.")
            return
        elif args[0] in ["upgrade", "u", "up", "study", "learn"]:
            agt = Agent().load(scp_cards.get(event, qid=qid))
            anb = agt.all_not_base()

            if args[1] in all_alias_dict.keys():
                key_name = all_alias_dict[args[1]]
                oldattr = getattr(agt, anb[key_name])
                level = int(oldattr[key_name])

                if len(args) <= 2:
                    up = level + 1
                else:
                    try:
                        up = int(args[2])
                    except ValueError:
                        await matcher.send(help_message("scp"))
                        return

                if level >= up:
                    await matcher.send(f"[Oracle] 你的 {args[1]} 技能的等级已经是 {level} 级了.")
                    return

                required = int(up * (up + 1) / 2 - level * (level + 1) / 2)
                if agt.p[anb[key_name]] < required:
                    await matcher.send(f"[Oracle] 你的熟练值不足以支撑你将 {args[1]} 提升到 {up} 级.")
                    return

                agt.p[anb[key_name]] -= required

                flt = random.randint(1, 10)

                if flt == 10:
                    flt = 0

                oldattr[key_name] = float(up) + flt/10
                setattr(agt, anb[key_name], oldattr)
                scp_cards.update(event, agt.__dict__, qid=qid, save=True)
                await matcher.send(f"[Oracle] 你的 {args[1]} 升级到 {up} 级.\n该技能的熟练度为 {oldattr[key_name]}.")
                return
            else:
                await matcher.send(f"[Oracle] 自定义技能 {args[1]} 无法被升级.")
                return
        elif args[0] in ["deal", "d", "buy", "b"]:
            args_for_deal = args[1:]
            await matcher.send(deal(event, args_for_deal))
            return 
        else:
            await matcher.send(f"[Oracle] 指令 {args[0]} 看起来似乎不存在.")
            return

    agt = Agent()
    agt.init()

    scp_cache_cards.update(event, agt.__dict__, save=False)
    await matcher.send(str(agt.output()))

@dndcommand.handle()
async def dnd_handler(matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(event.get_message(), begin=".dnd")
    if len(args) > 1:
        logger.info("指令错误, 驳回.")
        await matcher.send("[Oracle] 错误: 参数超出预计(1需要 但 %d传入), 指令驳回." % len(args))
        return True

    try:
        if len(args) == 0:
            raise ValueError
        args = int(args[0])
    except ValueError:
        await matcher.send(f'警告: 参数 {args} 不合法, 使用默认值 20 替代.')
        args = 20

    adv = Adventurer()
    adv.age_check(args)
    adv.init()
    
    if adv.int[0] <= 8:
        await matcher.send("[Orcale] 很遗憾, 检定新的冒险者智力不足, 弱智是不允许成为冒险者的, 请重新进行车卡检定.")
        return True

    if 15 <= args and args < 90:
        dnd_cache_cards.update(event, adv.__dict__, save=False)
        await matcher.send(str(adv.output()))
    return True

@showcommand.handle()
async def showhandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(event.get_message(), begin=(".show", ".display"))
    at = get_mentions(event)

    if mode in modes:
        try:
            sh = show_handler(event, args, at, mode=mode)
        except Exception as error:
            logger.exception(error)
            sh = [f"[Oracle] 错误: 执行指令失败, 疑似该模式不存在该指令."]
    else:
        await matcher.send("未知的跑团模式.")
        return True

    for msg in sh:
        await matcher.send(str(msg))
    return

@setcommand.handle()
async def sethandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(event.get_message(), begin=(".set", ".st"))
    at = get_mentions(event)

    if at and not is_super_user(event):
        await matcher.send("[Oracle] 权限不足, 拒绝执行指令.")
        return

    try:
        sh = set_handler(event, args, at, mode=mode)
    except Exception as error:
        logger.exception(error)
        sh = [f"[Oracle] 错误: 执行指令失败, 疑似模式 {mode} 不存在该指令."]

    await matcher.send(sh)
    return


@helpcommand.handle()
async def rdhelphandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(str(event.get_message()), begin=(".help", ".h"))
    if args:
        arg = args[0]
    else:
        arg = ""
    print(arg)
    await matcher.send(help_message(arg))


@modecommand.handle()
async def modehandler(matcher: Matcher, event: GroupMessageEvent):
    global mode
    args = format_msg(event.get_message(), begin=(".mode", ".m"))
    if args:
        if args[0].lower() in modes:
            mode = args[0].lower()
            await matcher.send(f"[Oracle] 已切换到 {mode.upper()} 跑团模式.")
            return True
        else:
            await matcher.send("[Oracle] 未知的跑团模式, 忽略.")
            await matcher.send(help_message("mode"))
            return True
    else:
        await matcher.send(f"[Oracle] 当前的跑团模式为 {mode.upper()}.")

@stcommand.handle()
async def stcommandhandler(matcher: Matcher, event: GroupMessageEvent):
    try:
        await matcher.send(st())
    except:
        await matcher.send(help_message("st"))


@attackcommand.handle()
async def attackhandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_str(event.get_message(), begin=(".at", ".attack"))
    if mode == "coc":
        await matcher.send(at(args, event))
    elif mode == "scp":
        await matcher.send(sat(args, event))


@damcommand.handle()
async def damhandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(event.get_message(), begin=(".dam", ".damage"))
    if mode == "scp":
        sd = scp_dam(args, event)
    elif mode == "coc":
        sd = coc_dam(args, event)
    else:
        await matcher.send("未知的跑团模式.")
        return

    await matcher.send(sd)

@encommand.handle()
async def enhandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(event.get_message(), begin=".en")
    at = get_mentions(event)

    if at and not is_super_user(event):
        await matcher.send("[Oracle] 权限不足, 拒绝执行指令.")
        return

    try:
        en = eval(f"{mode}_en(event, args)")
    except:
        en = f"[Oracle] 错误: 执行指令失败, 疑似模式 {mode.upper()} 不存在该指令."

    await matcher.send(en)


@racommand.handle()
async def rahandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(event.get_message(), begin=".ra")
    if mode in ["coc", "scp", "dnd"]:
        if mode == "scp":
            await matcher.send(sra(args, event))
        elif mode == "coc":
            await matcher.send(ra(args, event))
        elif mode == "dnd":
            await matcher.send(dra(args, event))
    return

@rhcommand.handle()
async def rhhandler(bot: Bot, matcher: Matcher, event: GroupMessageEvent):
    args = format_str(event.get_message(), begin=".rh")
    await matcher.send("[Oracle] 暗骰: 命运的骰子在滚动.")
    await bot.send_private_msg(user_id=event.get_user_id(), message=roll(args))

@rhacommand.handle()
async def rhahandler(bot: Bot, matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(event.get_message(), begin=".rha")
    await matcher.send("[Oracle] 暗骰: 命运的骰子在滚动.")
    await bot.send_private_msg(user_id=event.get_user_id(), message=ra(args, event))

@rcommand.handle()
async def rollhandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_msg(event.get_message(), begin=(".r", ".roll"))
    if not args:
        await matcher.send(roll(["1", "d", "100"]))
        return

    if args[0] == "b":
        await matcher.send(rb(args[1:]))
        return
    elif args[0] == "p":
        await matcher.send(rp(args[1:]))
        return

    try:
        await matcher.send(roll(args))
    except:
        await matcher.send(help_message("r"))


@ticommand.handle()
async def ticommandhandler(matcher: Matcher, event: GroupMessageEvent):
    try:
        await matcher.send(ti())
    except:
        await matcher.send(help_message("ti"))


@licommand.handle()
async def licommandhandler(matcher: Matcher, event: GroupMessageEvent):
    try:
        await matcher.send(li())
    except:
        await matcher.send(help_message("li"))


@sccommand.handle()
async def schandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_str(event.get_message(), begin=".sc")
    scrs = sc(args, event)

    if isinstance(scrs, list):
        for scr in scrs:
            await matcher.send(scr)
    else:
        await matcher.send(scrs)

@delcommand.handle()
async def delhandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_str(event.get_message(), begin=(".del", ".delete"))
    at = get_mentions(event)

    if at and not is_super_user(event):
        await matcher.send("[Oracle] 权限不足, 拒绝执行指令.")
        return

    if mode in modes:
        for msg in del_handler(event, args, at, mode=mode):
            await matcher.send(msg)

@chatcommand.handle()
async def chathandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_str(event.get_message(), begin=".chat")
    if not args:
        await matcher.send("[Oracle] 空消息是不被允许的.")
        return
    await matcher.send(chat(args))


@versioncommand.handle()
async def versionhandler(matcher: Matcher, event: GroupMessageEvent):
    args = format_str(event.get_message(), begin=(".version", ".v"))
    await matcher.send(f"欧若可骰娘 版本 {version}, 未知访客开发, 以Apache-2.0协议开源.\nCopyright © 2011-2023 Unknown Visitor. Open source as protocol Apache-2.0.")
    return

class OracleClient(botpy.Client):
    handlers = get_handlers(__import__(__name__))

    async def on_ready(self):
        global DEBUG
        if DEBUG:
            get_logger().setLevel(logging.DEBUG)
            logger.remove()
            logger.add(
                sys.stdout,
                level = "DEBUG"
            )
            logger.info("DEBUG 模式已启动.")
        init()
        coc_cards.load()
        scp_cards.load()
        dnd_cards.load()
        logger.success("机器人启动成功, 将进行心跳维持链接.")

    async def on_at_message_create(self, message: Message):
        is_command = False
        for handler in self.handlers:
            if await handler(api=self.api, message=message, params=None):
                isbot = "玩家" if message.author.bot == False else "机器人"
                logger.info(f"[dicer] 执行指令: {message.content} 指令来源: {message.channel_id} : {message.author.id} : {message.author.username} : {isbot}")
                is_command = True
                break
        valid = message.content.startswith(".") and len(message.content) >= 2
        if not is_command and valid:
            await message.reply(content="[Oracle] 不是合格的指令, 请检查你的输入.")

    async def on_message_create(self, message: Message):
        is_command = False
        if message.mentions:
            return
        for handler in self.handlers:
            if await handler(api=self.api, message=message, params=None):
                isbot = "玩家" if message.author.bot == False else "机器人"
                logger.info(f"[dicer] 执行指令: {message.content} 指令来源: {message.channel_id} : {message.author.id} : {message.author.username} : {isbot}")
                is_command = True
                break
        valid = message.content.startswith(".") and len(message.content) >= 2
        if not is_command and valid:
            await message.reply(content="[Oracle] 不是合格的指令, 请检查你的输入.")

def main():
    intents = botpy.Intents.all()
    client = OracleClient(intents=intents)
    run = lambda: client.run(appid=config["appid"], token=config["token"])
    logger.info("启动`QQGuild`机器人服务中...")
    run()

if __name__ == "__main__":
    main()



class NoneTypeCommandError(ValueError):
    """ 预留指令槽位空错误 """
    ...

class CommandRequired(TypeError):
    """ 必须参数未传入错误 """
    ...

class TooManyAliasCommandError(TypeError):
    """ 同义参数传入多个错误 """
    ...


from .trailblazer import Trailblazer
from .nbhandlers import commands
from .hsrcards import hsr_cards

hsr_cards.load()

__type__ = "plugin"
__charactor__ = Trailblazer
__name__ = "hsr"
__cname__ = "开拓者"
__nbhandler__ = nbhandlers
__nbcommands__ = commands


from loguru._logger import Logger, Core
from typing import Union
import sys
import re

def multilogger(
        sink = sys.stdout,
        name: str = "",
        payload: str = "",
        format: str = "<!time>[<level>{level}</level>] <cyan><!name></cyan> | <!payload><!module><level>{message}</level>",
        colorize: bool = True,
        level: str = "INFO",
        notime: bool = False,
        *args,
        **kwargs
) -> Logger:
    """ 日志管理系统 """
    module = "" if level != "DEBUG" else "<cyan>{module}</cyan>.<cyan>{name}</cyan>:{line} | "
    payload = f"<red>{payload}</red> | " if payload else ""
    time = "<green>{time:YYYY-MM-DD HH:mm:ss}</green> " if not notime else ""

    for match in re.findall(r"(<!.*?>)", format):
        value = re.match(r"^<!(.*?)>$", match)[1]
        format = re.sub(match, eval(value), format)

    logger_instance = Logger(
        core=Core(),
        exception=None,
        depth=0,
        record=False,
        lazy=False,
        colors=False,
        raw=False,
        capture=True,
        patchers=[],
        extra={},
    )
    logger_instance.configure(handlers=[
        {
            "sink": sink,
            "format": format,
            "colorize": colorize,
            "level": level,
        },
    ])
    return logger_instance

if __name__ == "__main__":
    log = multilogger()
    log.info("hhh")
    log2 = multilogger(payload="ARP", notime=False)
    log2.info("hhhhh")
    log.info("hh")


from .agent import Agent
from .scpcards import scp_cards, scp_cache_cards
from .scputils import deal, begin
from .attributes import all_alias_dict

try:
    from dicergirl.utils.utils import format_msg, get_status, get_mentions, is_super_user, on_startswith
    from dicergirl.utils.parser import CommandParser, Commands, Only, Optional, Required
except ImportError:
    from ..utils.utils import format_msg, get_status, get_mentions, is_super_user, on_startswith
    from ..utils.parser import CommandParser, Commands, Only, Optional, Required

from nonebot.matcher import Matcher
from nonebot.adapters import Bot as Bot
from nonebot.adapters.onebot.v11 import Bot as V11Bot
from nonebot.internal.matcher.matcher import Matcher
from nonebot.adapters.onebot.v11 import GroupMessageEvent

import asyncio
import random

scpcommand = on_startswith(".scp", priority=1, block=True).handle()

async def scp_handler(matcher: Matcher, event: GroupMessageEvent):
    """ SCP 车卡指令 """
    if not get_status(event):
        return

    args = format_msg(event.get_message(), begin=".scp", zh_en=True)
    at = get_mentions(event)

    if at and not is_super_user(event):
        return "[Oracle] 权限不足, 无法指定玩家修改人物卡."

    if at:
        qid = at[0]
    else:
        qid = ""

    if len(args) != 0:
        if args[0] in ["begin", "start"]:
            for des in begin():
                await matcher.send(des)
                await asyncio.sleep(2)
            return
        elif args[0] in ["level", "levelup", "lu"]:
            agt = Agent().load(scp_cards.get(event, qid=qid))

            if agt.ach >= 10:
                agt.ach -= 10
                agt.level += 1
                scp_cards.update(event, inv_dict=agt.__dict__, save=True)
                await matcher.send(f"[Oracle] {agt.name} 特工权限提升至 {agt.level} 级.")
                return
            else:
                await matcher.send(f"[Oracle] 特工 {agt.name} 功勋不足, 无法申请提升权限等级.")
                return
        elif args[0] in ["reset", "r"]:
            if not is_super_user(event):
                await matcher.send("[Oracle] 权限不足, 拒绝执行人物卡重置指令.")
                return

            agt = Agent().load(scp_cards.get(event, qid=qid))

            if len(args) == 2:
                try:
                    exec(f"agt.reset_{args[1]}()")
                    scp_cards.update(event, agt.__dict__, qid=qid, save=True)
                    await matcher.send(f"[Oracle] 已重置指定人物卡属性: {args[1]}.")
                except:
                    await matcher.send("[Oracle] 指令看起来不存在.")
                finally:
                    return

            agt.reset()
            scp_cards.update(event, agt.__dict__, qid=qid, save=True)
            await matcher.send(f"[Oracle] 人物卡 {agt.name} 属性已重置.")
            return
        elif args[0] in ["upgrade", "u", "up", "study", "learn"]:
            agt = Agent().load(scp_cards.get(event, qid=qid))
            anb = agt.all_not_base()

            if args[1] in all_alias_dict.keys():
                key_name = all_alias_dict[args[1]]
                oldattr = getattr(agt, anb[key_name])
                level = int(oldattr[key_name])

                if len(args) <= 2:
                    up = level + 1
                else:
                    try:
                        up = int(args[2])
                    except ValueError:
                        await matcher.send("[Oracle] 需要提升的等级应当为整型数, 请检查你的指令.\n使用`.help scp`查看 SCP 车卡指令使用方法.")
                        return

                if level >= up:
                    await matcher.send(f"[Oracle] 你的 {args[1]} 技能的等级已经是 {level} 级了.")
                    return

                required = int(up * (up + 1) / 2 - level * (level + 1) / 2)
                if agt.p[anb[key_name]] < required:
                    await matcher.send(f"[Oracle] 你的熟练值不足以支撑你将 {args[1]} 提升到 {up} 级.")
                    return

                agt.p[anb[key_name]] -= required

                flt = random.randint(1, 10)

                if flt == 10:
                    flt = 0

                oldattr[key_name] = float(up) + flt/10
                setattr(agt, anb[key_name], oldattr)
                scp_cards.update(event, agt.__dict__, qid=qid, save=True)
                await matcher.send(f"[Oracle] 你的 {args[1]} 升级到 {up} 级.\n该技能的熟练度为 {oldattr[key_name]}.")
                return
            else:
                await matcher.send(f"[Oracle] 自定义技能 {args[1]} 无法被升级.")
                return
        elif args[0] in ["deal", "d", "buy", "b"]:
            args_for_deal = args[1:]
            await matcher.send(deal(event, args_for_deal))
            return 
        else:
            await matcher.send(f"[Oracle] 指令 {args[0]} 看起来似乎不存在.")
            return

    agt = Agent()
    agt.init()

    scp_cache_cards.update(event, agt.__dict__, save=False)
    await matcher.send(str(agt.output()))

commands = {"scpcommand": "scp_handler"}


from pathlib import Path
from multilogging import multilogger

try:
    from .utils import get_group_id, get_user_id
except ImportError:
    from dicergirl.utils.utils import get_group_id, get_user_id

import json

logger = multilogger(name="Dicer Girl", payload="Card")

class Cards():
    data_path: Path = Path.home() / ".dicergirl" / "data"

    def __init__(self, mode: str=None, cache_path: Path=None):
        self.data = {}
        self.mode = mode if mode else "未知模式"
        self.cache_path = cache_path if cache_path else self.data_path / f"{mode}_cards.json"

    def init(self):
        logger.info(f"{self.mode.upper()} 存储文件未建立, 建立它.")
        with open(self.cache_path, "w", encoding="utf-8") as f:
            json.dump({}, f, ensure_ascii=False)

    def save(self):
        logger.info(f"保存 {self.mode.upper()} 人物卡数据.")
        with open(self.cache_path, "w", encoding="utf-8") as f:
            json.dump(self.data, f, ensure_ascii=False)

    def load(self) -> dict:
        if not self.cache_path.exists():
            self.init()

        with open(self.cache_path, "r", encoding="utf-8") as f:
            data = f.read()
            if not data:
                self.data = {}
            else:
                self.data = json.loads(data)

        return self.data

    def update(self, message, inv_dict, qid="", save=True):
        group_id = get_group_id(message)
        if not self.data.get(group_id):
            self.data[group_id] = {}
        self.data[group_id].update(
            {qid if qid else get_user_id(message): inv_dict}
            )
        if save:
            self.save()

    def get(self, message, qid=""):
        group_id = get_group_id(message)
        if self.data.get(group_id):
            if self.data[group_id].get(qid if qid else get_user_id(message)):
                return self.data[group_id].get(qid if qid else get_user_id(message))
        else:
            return None

    def delete(self, message, qid: str = "", save: bool = True) -> bool:
        if self.get(message, qid=qid):
            if self.data[get_group_id(message)].get(qid if qid else get_user_id(message)):
                self.data[get_group_id(message)].pop(
                    qid if qid else get_user_id(message))
            if save:
                self.save()
            return True
        return False

    def delete_skill(self, message, skill_name: str, qid: str = "", save: bool = True) -> bool:
        if self.get(message, qid=qid):
            data = self.get(message, qid=qid)
            if data["skills"].get(skill_name):
                data["skills"].pop(skill_name)
                self.update(message, data, qid=qid, save=save)
                return True
        return False


from pathlib import Path
from typing import List, Dict
try:
    from dicergirl.utils.multilogging import multilogger
except ImportError:
    from .multilogging import multilogger

import sys
import yaml

package: str = None
""" 目前 Dicer Girl 挂载的平台 """
allowed_packages: List[str] = ["nonebot2", "qqguild"]
""" Dicer Girl 允许挂载的平台 """
logger = multilogger(name="Dicer Girl", payload="Settings")
""" `settings.py`日志管理系统 """
status: Dict[str, bool] = {}
""" 机器人在各个群聊状态 """

def set_package(pkg: str):
    """ 设置当前挂载的平台 """
    global package
    pkg = pkg.lower()

    if pkg in allowed_packages:
        package = pkg
        return package
    else:
        try:
            raise ValueError(f"错误的包名:`{pkg}`, 支持的包: {allowed_packages}")
        except Exception as error:
            logger.exception(error)
        sys.exit()

def get_package():
    """ 获得当前挂载的平台 """
    return package

def setconfig(appid, token, path=Path.home()/".dicergirl", filename="config.yaml"):
    """ 在`QQGuild`模式中设置频道机器人`appid`以及`token`. """
    if package == "nonebot2":
        raise AttributeError("你无法在 Nonebot2 模式下创建配置文件, 请确保你在调用`setconfig`函数之前已经执行了`set_package('qqguild').")
    configfile = open(path / filename, "w")
    data = {
            "appid": appid,
            "token": token
            }
    yaml.dump(data, configfile, sort_keys = False)
    return data

def getconfig(path=Path.home()/".dicergirl", filename="config.yaml"):
    """ 获取`QQGuild`模式中频道机器人的`appid`以及`token`. """
    configfile = open(path / filename, "r")
    config = yaml.safe_load(configfile.read())
    return config

def change_status(var) -> bool:
    """ 录入新的`status`内容 """
    global status
    status = var
    return True

def load_status_settings():
    """ 导出当前机器人在各群聊的状态 """
    return status

if __name__ == "__main__":
    set_package("?")


import random

build_dict = {64: -2, 84: -1, 124: 0, 164: 1,
              204: 2, 284: 3, 364: 4, 444: 5, 524: 6}
db_dict = {-2: "-2", -1: "-1", 0: "0", 1: "1d4",
           2: "1d6", 3: "2d6", 4: "3d6", 5: "4d6", 6: "5d6"}

def randattr(time: int = 3, ex: int = 0):
    r = 0
    for _ in range(time):
        r += random.randint(1, 6)
    return (r+ex)*5

class Investigator(object):
    """ COC 调查员人物卡 """
    def __init__(self) -> None:
        self.name = "无名调查员"
        self.age = 20
        self.sex = "女"
        self.str = randattr()
        self.con = randattr()
        self.siz = randattr(2, 6)
        self.dex = randattr()
        self.app = randattr()
        self.int = randattr(2, 6)
        self.pow = randattr()
        self.edu = randattr(2, 6)
        self.luc = randattr()
        self.san = self.pow
        self.hp = self.lp_max()
        self.skills = {}
        self.tools = {}

    def body_build(self) -> int:
        build = self.str + self.con
        for i, j in build_dict.items():
            if build <= i:
                return j
        return

    def db(self) -> str:
        return db_dict[self.body_build()]

    def lp_max(self) -> int:
        return (self.con+self.siz)//10

    def mov(self) -> int:
        r = 8
        if self.age >= 80:
            r -= 5
        elif self.age >= 70:
            r -= 4
        elif self.age >= 60:
            r -= 3
        elif self.age >= 50:
            r -= 2
        elif self.age >= 40:
            r -= 1
        if self.str < self.siz and self.dex < self.siz:
            return r-1
        elif self.str > self.siz and self.dex > self.siz:
            return r+1
        else:
            return r

    def edu_up(self) -> str:
        edu_check = random.randint(1, 100)
        if edu_check > self.edu:
            edu_en = random.randint(1, 10)
            self.edu += edu_en
        else:
            return "教育成长检定D100=%d, 小于%d, 无增长。" % (edu_check, self.edu)
        if self.edu > 99:
            self.edu = 99
            return "教育成长检定D100=%d, 成长1D10=%d, 成长到了最高值99！" % (edu_check, edu_en)
        else:
            return "教育成长检定D100=%d, 成长1D10=%d, 成长到了%d" % (edu_check, edu_en, self.edu)

    def edu_ups(self, times) -> str:
        r = ""
        for _ in range(times):
            r += self.edu_up()
        return r

    def sum_down(self, sum) -> str:
        if self.str + self.con + self.dex-45 < sum:
            self.str = 15
            self.con = 15
            self.dex = 15
        else:
            str_lost = random.randint(0, min(sum, self.str-15))
            while sum - str_lost > self.con + self.dex-30:
                str_lost = random.randint(0, min(sum, self.str-15))
            self.str -= str_lost
            sum -= str_lost
            con_lost = random.randint(0, min(sum, self.con-15))
            while sum - con_lost > self.dex-15:
                con_lost = random.randint(0, min(sum, self.con-15))
            self.con -= con_lost
            sum -= con_lost
            self.dex -= sum
        return

    def age_change(self, age: int = 20) -> str:
        if self.age != 20:
            return  # 防止多次年龄增强判定
        if age < 15:
            return "年龄过小, 无法担当调查员"
        elif age >= 90:
            return "该调查员已经作古, 他(或者她)由于年龄过大, 遭到了神祇「克苏鲁」的注视."
        self.age = age
        if 15 <= age < 20:
            self.str -= 5
            self.siz -= 5
            self.edu -= 5
            luc = randattr()
            self.luc = luc if luc > self.luc else self.luc
            return "力量、体型、教育值-5, 幸运增强判定一次"
        elif age < 40:
            self.edu_up()
            return "教育增强判定一次"
        elif age < 50:
            self.app -= 5
            self.sum_down(5)
            self.edu_ups(2)
            return "外貌-5, 力量、体型、敏捷合计降低5, 教育增强判定两次"
        elif age < 60:
            self.app -= 10
            self.sum_down(10)
            self.edu_ups(3)
            return "外貌-10, 力量、体型、敏捷合计降低10, 教育增强判定三次"
        elif age < 70:
            self.app -= 15
            self.sum_down(20)
            self.edu_ups(4)
            return "外貌-15, 力量、体型、敏捷合计降低20, 教育增强判定四次"
        elif age < 80:
            self.app -= 20
            self.sum_down(40)
            self.edu_ups(4)
            return "外貌-20, 力量、体型、敏捷合计降低40, 教育增强判定四次"
        elif age < 90:
            self.app -= 25
            self.sum_down(80)
            self.edu_ups(4)
            return "外貌-25, 力量、体型、敏捷合计降低80, 教育增强判定四次"

    def __repr__(self):
        data = "姓名: %s\n" % self.name
        data += "性别: %s 年龄: %d\n" % (self.sex, self.age)
        data += "力量: %d 体质: %d 体型: %d\n" % (self.str, self.con, self.siz)
        data += "敏捷: %d 外貌: %d 智力: %d\n" % (self.dex, self.app, self.int)
        data += "意志: %d 教育: %d 幸运: %d\n" % (self.pow, self.edu, self.luc)
        data += "DB: %s 移动速度: %d SAN: %d\n" % (self.db(), self.mov(), self.san)
        data += "生命值: %d/%d" % (self.hp, self.lp_max())
        return data

    def skills_output(self) -> str:
        if not self.skills:
            return "%s 当前无任何技能数据。" % self.name
        r = "%s技能数据: " % self.name
        for k, v in self.skills.items():
            r += "\n%s: %d" % (k, v)
        return r

    def output(self) -> str:
        return self.__repr__()

    def rollcount(self) -> tuple:
        return (self.__count(), self.__count() + self.luc)

    def __count(self):
        return self.str + self.con + self.siz + self.dex + self.app + self.int + self.pow + self.edu

    def load(self, data: dict):
        self.__dict__.update(data)
        return self



from .investigator import Investigator
from .nbhandlers import commands
from .coccards import coc_cards, coc_cache_cards, coc_attrs_dict
from .cocutils import coc_at, coc_dam, coc_ra, coc_en

coc_cards.load()

__type__ = "plugin"
__charactor__ = Investigator
__name__ = "coc"
__cname__ = "调查员"
__cards__ = coc_cards
__cache__ = coc_cache_cards
__nbhandler__ = nbhandlers
__nbcommands__ = commands
__commands__ = {
    "at": coc_at,
    "dam": coc_dam,
    "ra": coc_ra,
    "en": coc_en
}
__baseattrs__ = coc_attrs_dict


try:
    from ..utils.docimasy import expr, scp_doc
    from ..utils.dicer import Dicer
    from .scpcards import scp_cards
    from .attributes import all_names, scp_attrs_dict as attrs_dict, weapons, all_alias, all_alias_dict
    from .agent import Agent
    from ..utils.multilogging import multilogger
except ImportError:
    from dicergirl.utils.docimasy import expr, scp_doc
    from dicergirl.utils.dicer import Dicer
    from dicergirl.scp.scpcards import scp_cards
    from dicergirl.scp.attributes import all_names, scp_attrs_dict as attrs_dict, weapons, all_alias, all_alias_dict
    from dicergirl.scp.agent import Agent
    from dicergirl.utils.multilogging import multilogger

import random

logger = multilogger(name="Dicer Girl", payload="SCPUtil")

def scp_at(event, args):
    """ SCP 伤害检定 """
    card = scp_cards.get(event)
    agt = Agent().load(card)
    all_dices = []

    if not args:
        dices = [dice for dice in agt.dices["str"]]

        if len(dices) > 4:
            while len(all_dices) != 4:
                choice = random.choice(dices)
                all_dices.append(choice)
                dices.remove(choice)
        elif len(dices) <= 4:
            all_dices = dices

        results = []
        for dice in all_dices:
            dice = Dicer("1"+dice.lower()).roll()
            results.append(dice.outcome)

        result = max(results)

        if len(results) > 1:
            results.remove(result)
            result += max(results)

        return f"[Oracle] 特工发起近战格斗伤害检定, 检定造成了 {result} 点 伤害."
    else:
        args = "".join(args)

        upper = {name.upper(): [tool, name] for name, tool in agt.tools.items()}
        if not args.upper() in upper.keys():
            return f"[Oracle] 看起来该特工并未购置 {args.upper()}."

        return f"[Oracle] 特工使用 {upper[args.upper()][1]} 发起攻击, 检定造成了 {Dicer(upper[args.upper()][0]['base']).roll().calc()} 点 伤害."

def deal(event, args):
    """ SCP 武器交易系统 """
    if len(args) > 0:
        args = "".join(args).upper()

    card = scp_cards.get(event)
    level = card["level"]
    reply = ""

    if not args:
        reply += f"特工权限: {level}\n"

        for lvl in range(level):
            reply += f"Level {lvl+1} 准允购置的装备:"
            for weapon in weapons[lvl+1].keys():
                reply += f"\n  {weapon}: {weapons[lvl+1][weapon]['price']}￥"

            reply += "\n"

        return reply
    
    allowed_upper = {}
    for lvl in weapons.keys():
        allowed_upper.update({allow.upper(): [lvl, weapon, allow] for allow, weapon in weapons[lvl].items()})

    if args in allowed_upper.keys():
        real_name = allowed_upper[args][2]
        if allowed_upper[args][1]['price'] > card['money']:
            return f"[Oracle] 该特工囊中羞涩, 无法购置装备 {args}.\n贫穷是人类社会生存中的重大危机, 很高兴, 你距离危险更近了一步."

        card['money'] -= allowed_upper[args][1]['price']
        agt = Agent().load(card)
        agt.tools[real_name] = allowed_upper[args][1]
        scp_cards.update(event, agt.__dict__, save=True)
        return f"[Oracle] 特工购置了 1 件 {real_name}."
    else:
        return f"[Oracle] 装备 {real_name} 不存在或特工权限不足."

def scp_dam(event, args):
    """ SCP 承伤检定 """
    card = scp_cards.get(event)

    if not card:
        return "[Oracle] 未找到缓存数据, 请先使用`.scp`指令进行车卡生成角色卡并`.set`进行保存."

    max_hp = card["hp_max"]

    if len(args) == 1:
        if not args[0] in ["check", "c"]:
            arg = int(args[0])
            card["hp"] -= arg
            r = f"[Orcale] {card['name']} 失去了 {arg}点 生命"
        else:
            r = "检查特工状态"
    elif len(args) == 0:
        d = Dicer().parse("1d6").roll()
        card["hp"] -= d.outcome
        r = "[Oracle] 投掷 1D6={d}\n受到了 {d}点 伤害".format(d=d.calc())
    elif len(args) == 3:
        if args[1] != "d":
            r = "[Oracle] 未知的指令格式."
        else:
            d = Dicer().parse(f"{args[0]}{args[1]}{args[2]}").roll()
            card["hp"] -= d.outcome
            r = f"[Oracle] 投掷 {args[0]}D{args[2]}={d.calc()}\n受到了 {d.calc()}点 伤害"

    if card["hp"] <= 0:
        r += f", 特工 {card['name']} 已死亡."
    elif (max_hp * 0.8 <= card["hp"]) and (card["hp"] < max_hp):
        r += f", 特工 {card['name']} 具有轻微伤势."
    elif (max_hp * 0.6 <= card["hp"]) and (card['hp'] <= max_hp * 0.8):
        r += f", 特工 {card['name']} 具有轻微伤."
    elif (max_hp * 0.4 <= card["hp"]) and (card["hp"] <= max_hp * 0.6):
        r += f", 特工 {card['name']} 具有轻伤."
    elif (max_hp * 0.2 <= card["hp"]) and (card["hp"] <= max_hp * 0.4):
        r += f", 特工 {card['name']} 身负重伤."
    elif max_hp * 0.2 >= card["hp"]:
        r += f", 特工 {card['name']} 濒死."
    else:
        r += "."

    scp_cards.update(event, card)
    return r

def scp_ra(event, args: list) -> str:
    """ SCP 属性或技能检定 """
    if len(args) == 0:
        return "[Oracle] 错误: 检定技能需要给入技能名称.\n使用`.help ra`指令查看指令使用方法."
    elif len(args) > 4:
        return "[Oracle] 错误: 参数过多(最多4需要但%d给予).\n使用`.help ra`指令查看指令使用方法." % len(args)

    try:
        difficulty = int(args[-1])
        args.remove(args[-1])
    except ValueError:
        difficulty = 12

    card_data = scp_cards.get(event)

    if not card_data:
        return "[Oracle] 在执行参数检定前, 请先执行`.scp`车卡并执行`.set`保存."

    inv = Agent().load(card_data)

    is_base = False
    if len(args) == 1:
        for alias in attrs_dict.values():
            if args[0] in alias:
                dices = [dice for dice in inv.dices[alias[0]]]
                to_ens = [alias[0]]
                is_base = True
                break

    is_skill = False
    skill_only = False
    if not is_base and len(args) == 3:
        if args[1] in ["+", "/", "&", "*"]:
            is_validated_skill = False
            for alias in attrs_dict.values():
                if args[0] in alias:
                    dices = [dice for dice in inv.dices[alias[0]]]
                    to_ens = [alias[0]]
                    is_validated_skill = True
                    break

            anb = inv.all_not_base()
            if args[2] in anb.keys() and is_validated_skill:
                skill_name = [args[2], anb[args[2]]]
                exp = getattr(inv, anb[args[2]])[args[2]]
                is_skill = True
            elif not is_validated_skill:
                return f"[Oracle] 错误: 基础属性 {args[0]} 不存在."
            else:
                return f"[Oracle] 错误: 技能、知识或能力 {args[2]} 不存在."
        else:
            return "[Oracle] 我无法解析你的语法, 你可以使用`.help ra`指令查看指令使用方法.\n如果你确信这是一个错误, 建议联系开发者获得更多帮助.\n如果你是具有管理员权限, 你可以使用`.debug on`获得更多信息."
    elif not is_base and len(args) == 1:
        if args[0] in all_alias:
            anb = inv.all_not_base()
            key_name = all_alias_dict[args[0]]
            exp = getattr(inv, anb[key_name])[key_name]
            skill_only = True
            if anb[key_name] == "knowledge":
                to_ens = ["int", "per"]
                to_en = random.choice(to_ens)
                dices = [dice for dice in (inv.dices[to_en])]
            elif anb[key_name] == "skills":
                to_ens = ["str", "dex"]
                to_en = random.choice(to_ens)
                dices = [dice for dice in (inv.dices[to_en])]
            elif anb[key_name] == "ability":
                to_ens = ["chr", "wil"]
                to_en = random.choice(to_ens)
                dices = [dice for dice in (inv.dices[to_en])]
            else:
                skill_only = False
            skill_name = [key_name, anb[key_name]]

    if not is_base and not is_skill and not skill_only:
        if args[0] in inv.skills.keys():
            exp = inv.skills[args[0]]
            return str(expr(Dicer(), int(exp)))
        else:
            return "[Oracle] 错误: 没有这个数据或技能."

    all_dices = []

    if len(dices) > 4:
        while len(all_dices) != 4:
            choice = random.choice(dices)
            all_dices.append(choice)
            dices.remove(choice)
    elif len(dices) <= 4:
        all_dices = dices

    results = []
    great = False
    for dice in all_dices:
        dice = Dicer("1"+dice.lower()).roll()
        results.append(dice.outcome)
        if dice.great:
            great = True

    result = max(results)

    if len(results) > 1:
        results.remove(result)
        result += max(results)

    if is_skill or skill_only:
        result += exp

    encourage = None
    for en in card_data["en"]:
        if en in to_ens:
            encourage = card_data["en"][en]
            card_data["en"].pop(en)
            scp_cards.update(event, card_data)
            break

    r = scp_doc(result, difficulty, encourage=encourage, agent=inv.name, great=great)

    if is_skill or skill_only:
        if r.judge == 1:
            card_data["p"][skill_name[1]] += 0.1
        elif r.judge > 1:
            card_data["p"][skill_name[1]] += 1
        scp_cards.update(event, card_data)

    return r.detail

def scp_en(event, args):
    """ SCP 属性激励 """
    if not args:
        return "[Oracle] 错误: 你没有给定需要激励的基础属性.\n使用`.help en`指令查看指令使用方法."

    try:
        en = int(args[1])
        if not en:
            return f"[Oracle] 无法进行发起 {en} 点激励."
    except ValueError:
        return "[Oracle] 错误: 给定需要消耗的激励点应当为整型数.\n使用`.help en`指令查看指令使用方法."

    card_data = scp_cards.get(event)

    if card_data["enp"] < en:
        return f"[Oracle] 你仅剩的激励点无法进行发起 {en} 点激励."
    
    agt = Agent().load(card_data)

    for alias in attrs_dict.values():
        if args[0] in alias:
            to_en = alias[0]
            is_validated_skill = True
            break
    
    if not is_validated_skill:
        return f"[Oracle] 不存在的基础属性 {args[0]} 无法被激励."

    agt.enp -= en
    agt.en[to_en] = en
    scp_cards.update(event, agt.__dict__, save=True)

    return f"[Oracle] 你的 {args[0]} 受到了 {en} 点激励."

def begin():
    """ SCP 起始语 """
    reply = []
    reply.append("人类到如今已经繁衍了 250000 年, 只有最近的4000 年是有意义的.")
    reply.append("所以, 我们在将近 250000 年中在干什么? ")
    reply.append("我们躲在山洞中, 围坐在小小的篝火边, 畏惧那些我们不懂得的事物——那些关于太阳如何升起的解释, 那些人头鸟身的怪物, 那些有生命的石头.")
    reply.append("所以我们称他们为“神”和“恶魔”, 并向他们祈求宽恕和祈祷拯救. ")
    reply.append("之后, 他们的数量在减少, 我们的数量在增加.")
    reply.append("当我们恐惧的事物越来越少, 我们开始更理智的看待这个世界.")
    reply.append("然而, 不能解释的事物并没有消失, 就好像宇宙故意要表现出荒谬与不可思议一样.")
    reply.append("人类不能再生活在恐惧中.")
    reply.append("没有东西能保护我们, 我们必须保护我们自己.")
    reply.append("当其他人在阳光下生活时, 我们必须在阴影中和它们战斗, 并防止它们暴露在大众眼中, 这样其他人才能生活在一个理智的, 普通的世界中.")
    reply.append("我们控制, 我们收容, 我们保护.")
    reply.append("——The Administrator")
    return reply


from pathlib import Path
from multilogging import multilogger

import sys
import importlib

logger = multilogger(name="Dicer Girl", payload="Plugin")

loaded = False
""" 首次加载标识, 防止nb指令重复注册 """
def modules():
    global loaded
    if loaded:
        return modes

    modules_dict = {}
    sys.path.append(Path(__file__).resolve().parent.parent.__str__())

    for folder in Path(__file__).resolve().parent.parent.iterdir():
        if Path(folder).is_dir() and (Path(folder) / "__init__.py").exists():
            try:
                module = importlib.import_module(folder.name)
            except Exception as error:
                logger.exception(error)
                logger.error(f"插件 {folder.name} 导入失败.")
                continue

            if not hasattr(module, "__type__"):
                continue

            if module.__type__ != "plugin":
                continue

            if hasattr(module, "__nbcommands__"):
                commands: dict = module.__nbcommands__
            else:
                commands = {}

            if commands and not hasattr(module, "__nbhandler__"):
                logger.error(f"插件 {folder.name} 配置异常, 导入失败.")
                continue

            if hasattr(module, "__nbcommands__"):
                handlers = module.__nbhandler__
            else:
                handlers = {}

            for command, handler in commands.items():
                try:
                    getattr(handlers, command)(getattr(handlers, handler))
                except AttributeError:
                    logger.error(f"插件 {folder.name} 中 Nonebot2 指令配置异常, 导入失败.")
                    continue
                except Exception as error:
                    logger.error("未知错误:")
                    logger.exception(error)

            modules_dict[module.__name__] = module
            logger.success(f"插件 {folder.name.upper()} 导入完成.")

    sys.path.pop(-1)
    loaded = True
    return modules_dict

modules_dict = modules()
modes: dict = {module.split(".")[-1]: modules_dict[module] for module in modules_dict}
""" 已导入的跑团模块 """